#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Utilities common to all administration panels.
"""
from __future__ import unicode_literals
from django.contrib import admin


class BaseAdmin(admin.ModelAdmin):
    """
    Parent class with attributes common to all CAL-ACCESS data models.
    """
    save_on_top = True

    def get_readonly_fields(self, *args, **kwargs):
        """
        Make all fields read only.
        """
        return [f.name for f in self.model._meta.fields]

    def get_list_filter(self, request):
        """
        Filter all fields with `choices` configured.
        """
        return [f.name for f in self.model._meta.fields if f.choices]

    def get_search_fields(self, request):
        """
        Search all fields on the model.
        """
        return [f.name for f in self.model._meta.fields]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Custom managers for working with CAL-ACCESS data.
"""
from __future__ import unicode_literals
import os
from django.db import models
from calaccess_raw import get_download_directory


class CalAccessManager(models.Manager):
    """
    Utilities for accessing the raw data associated with a model.
    """
    def get_csv_name(self):
        """
        Returns the name of the model's CSV data file.
        """
        return "%s.csv" % self.model._meta.db_table.lower()

    def get_csv_path(self):
        """
        Returns the path to the model's CSV data file.
        """
        return os.path.join(
            get_download_directory(),
            'csv',
            self.get_csv_name()
        )

    def get_tsv_name(self):
        """
        Returns the name of the model's raw TSV data file.
        """
        return "%s.TSV" % self.model._meta.db_table

    def get_tsv_path(self):
        """
        Returns the path to the model's raw TSV data file.
        """
        return os.path.join(
            get_download_directory(),
            'tsv',
            self.get_tsv_name()
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Logs row count of given model and compares against line count in cleaned CSV.
"""
from __future__ import unicode_literals
from django.apps import apps
from django.db import router
from django.contrib.humanize.templatetags.humanize import intcomma
from calaccess_raw.management.commands import CalAccessCommand
from calaccess_raw.models.tracking import RawDataVersion


class Command(CalAccessCommand):
    """
    Logs row count of given model and compares against line count in cleaned CSV.
    """
    help = "Logs row count of given model and compares against line count in cleaned CSV"

    def add_arguments(self, parser):
        """
        Adds custom arguments specific to this command.
        """
        super(Command, self).add_arguments(parser)
        parser.add_argument('model_name', help="Name of model to verify")
        parser.add_argument(
            "-a",
            "--app-name",
            dest="app_name",
            default="calaccess_raw",
            help="Name of Django app with models into which data will "
                 "be imported (if other not calaccess_raw)"
        )

    def handle(self, *args, **options):
        """
        Make it happen.
        """
        super(Command, self).handle(*args, **options)

        if self.verbosity > 1:
            self.log(" Verifying %s" % options['model_name'])

        model = apps.get_model(options['app_name'], options['model_name'])

        self.database = router.db_for_write(model=model)

        # Get the model total
        model_count = model.objects.count()

        try:
            version = self.raw_data_versions.latest('release_datetime')
        except RawDataVersion.DoesNotExist:
            version = None
        else:
            raw_file = self.raw_data_files.get_or_create(
                version=version,
                file_name=model._meta.db_table
            )[0]

            # add load counts to raw_file_record
            raw_file.load_columns_count = len(model._meta.fields)
            raw_file.load_records_count = model_count
            raw_file.save()

        # Get the CSV total
        try:
            csv_count = raw_file.clean_records_count
        except UnboundLocalError:
            csv_path = model.objects.get_csv_path()
            with open(csv_path) as f:
                csv_count = sum(1 for l in f) - 1

        if self.verbosity > 1:
            # Report back on how we did
            if model_count == csv_count:
                self.success("  Record count matches CSV")
                self.success("  - Model: %s" % intcomma(model_count))
                self.success("  - CSV: %s" % intcomma(csv_count))
            else:
                self.failure("  Record count does not match CSV")
                self.failure("  - Model: %s" % intcomma(model_count))
                self.failure("  - CSV: %s" % intcomma(csv_count))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Import all of the admins from submodules and thread them together.
"""
from calaccess_raw.admin.base import BaseAdmin
from calaccess_raw.admin.campaign import (
    CvrSoCdAdmin,
    Cvr2SoCdAdmin,
    CvrCampaignDisclosureCdAdmin,
    Cvr2CampaignDisclosureCdAdmin,
    RcptCdAdmin,
    Cvr3VerificationInfoCdAdmin,
    LoanCdAdmin,
    S401CdAdmin,
    ExpnCdAdmin,
    F495P2CdAdmin,
    DebtCdAdmin,
    S496CdAdmin,
    S497CdAdmin,
    F501502CdAdmin,
    S498CdAdmin,
)
from calaccess_raw.admin.inactive import (
    BallotMeasuresCdAdmin,
    CvrF470CdAdmin,
    FilerTypePeriodsCd,
    LobbyistContributions1CdAdmin,
    LobbyistContributions2CdAdmin,
    LobbyistContributions3CdAdmin,
    LobbyistEmpLobbyist1CdAdmin,
    LobbyistEmpLobbyist2CdAdmin,
    LobbyistEmployer1CdAdmin,
    LobbyistEmployer2CdAdmin,
    LobbyistEmployer3CdAdmin,
    LobbyistEmployerFirms1CdAdmin,
    LobbyistEmployerFirms2CdAdmin,
    LobbyistEmployerHistoryCdAdmin,
    LobbyistFirm1CdAdmin,
    LobbyistFirm2CdAdmin,
    LobbyistFirm3CdAdmin,
    LobbyistFirmEmployer1CdAdmin,
    LobbyistFirmEmployer2CdAdmin,
    LobbyistFirmHistoryCdAdmin,
    LobbyistFirmLobbyist1CdAdmin,
    LobbyistFirmLobbyist2CdAdmin,
    EfsFilingLogCdAdmin,
)
from calaccess_raw.admin.lobbying import (
    CvrRegistrationCdAdmin,
    Cvr2RegistrationCdAdmin,
    CvrLobbyDisclosureCdAdmin,
    Cvr2LobbyDisclosureCdAdmin,
    LobbyAmendmentsCdAdmin,
    F690P2CdAdmin,
    LattCdAdmin,
    LexpCdAdmin,
    LccmCdAdmin,
    LothCdAdmin,
    LempCdAdmin,
    LpayCdAdmin,
    LobbyingChgLogCdAdmin
)
from calaccess_raw.admin.common import (
    FilernameCdAdmin,
    FilerFilingsCdAdmin,
    FilingsCdAdmin,
    SmryCdAdmin,
    CvrE530CdAdmin,
    SpltCdAdmin,
    TextMemoCdAdmin,
    AcronymsCdAdmin,
    AddressCdAdmin,
    FilersCdAdmin,
    FilerAcronymsCdAdmin,
    FilerAddressCdAdmin,
    FilerEthicsClassCdAdmin,
    FilerInterestsCdAdmin,
    FilerLinksCdAdmin,
    FilerStatusTypesCdAdmin,
    FilerToFilerTypeCdAdmin,
    FilerTypesCdAdmin,
    FilerXrefCdAdmin,
    FilingPeriodCdAdmin,
    GroupTypesCdAdmin,
    HeaderCdAdmin,
    HdrCdAdmin,
    ImageLinksCdAdmin,
    LegislativeSessionsCdAdmin,
    LookupCodesCdAdmin,
    NamesCdAdmin,
    ReceivedFilingsCdAdmin,
    ReportsCdAdmin,
)

from calaccess_raw.admin.tracking import (
    RawDataVersionAdmin,
    RawDataFileAdmin,
    RawDataCommandAdmin
)

__all__ = (
    'BaseAdmin',
    'CvrSoCdAdmin',
    'Cvr2SoCdAdmin',
    'CvrCampaignDisclosureCdAdmin',
    'Cvr2CampaignDisclosureCdAdmin',
    'RcptCdAdmin',
    'Cvr3VerificationInfoCdAdmin',
    'LoanCdAdmin',
    'S401CdAdmin',
    'ExpnCdAdmin',
    'F495P2CdAdmin',
    'DebtCdAdmin',
    'S496CdAdmin',
    'SpltCdAdmin',
    'S497CdAdmin',
    'F501502CdAdmin',
    'S498CdAdmin',
    'CvrF470CdAdmin',
    'CvrRegistrationCdAdmin',
    'Cvr2RegistrationCdAdmin',
    'CvrLobbyDisclosureCdAdmin',
    'Cvr2LobbyDisclosureCdAdmin',
    'LobbyAmendmentsCdAdmin',
    'F690P2CdAdmin',
    'LattCdAdmin',
    'LexpCdAdmin',
    'LccmCdAdmin',
    'LothCdAdmin',
    'LempCdAdmin',
    'LpayCdAdmin',
    'FilerFilingsCdAdmin',
    'FilingsCdAdmin',
    'SmryCdAdmin',
    'CvrE530CdAdmin',
    'TextMemoCdAdmin',
    'AcronymsCdAdmin',
    'AddressCdAdmin',
    'BallotMeasuresCdAdmin',
    'EfsFilingLogCdAdmin',
    'FilernameCdAdmin',
    'FilersCdAdmin',
    'FilerAcronymsCdAdmin',
    'FilerAddressCdAdmin',
    'FilerEthicsClassCdAdmin',
    'FilerInterestsCdAdmin',
    'FilerLinksCdAdmin',
    'FilerStatusTypesCdAdmin',
    'FilerToFilerTypeCdAdmin',
    'FilerTypesCdAdmin',
    'FilerXrefCdAdmin',
    'FilingPeriodCdAdmin',
    'FilerTypePeriodsCd',
    'GroupTypesCdAdmin',
    'HeaderCdAdmin',
    'HdrCdAdmin',
    'ImageLinksCdAdmin',
    'LegislativeSessionsCdAdmin',
    'LobbyingChgLogCdAdmin',
    'LobbyistContributions1CdAdmin',
    'LobbyistContributions2CdAdmin',
    'LobbyistContributions3CdAdmin',
    'LobbyistEmployer1CdAdmin',
    'LobbyistEmployer2CdAdmin',
    'LobbyistEmployer3CdAdmin',
    'LobbyistEmployerFirms1CdAdmin',
    'LobbyistEmployerFirms2CdAdmin',
    'LobbyistEmpLobbyist1CdAdmin',
    'LobbyistEmpLobbyist2CdAdmin',
    'LobbyistFirm1CdAdmin',
    'LobbyistFirm2CdAdmin',
    'LobbyistFirm3CdAdmin',
    'LobbyistFirmEmployer1CdAdmin',
    'LobbyistFirmEmployer2CdAdmin',
    'LobbyistFirmLobbyist1CdAdmin',
    'LobbyistFirmLobbyist2CdAdmin',
    'LobbyistFirmHistoryCdAdmin',
    'LobbyistEmployerHistoryCdAdmin',
    'LookupCodesCdAdmin',
    'NamesCdAdmin',
    'ReceivedFilingsCdAdmin',
    'ReportsCdAdmin',
    'RawDataVersionAdmin',
    'RawDataFileAdmin',
    'RawDataCommandAdmin',
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Configuration of the Django app.
"""
from __future__ import unicode_literals
from django.apps import AppConfig


class CalAccessRawConfig(AppConfig):
    """
    Configuration of the `calaccess_raw` Django app.
    """
    name = 'calaccess_raw'
    verbose_name = "CAL-ACCESS raw data"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Download, unzip, clean and load the latest CAL-ACCESS database ZIP.
"""
import os
import logging
from sys import exit
from hurry.filesize import size
from django.conf import settings
from clint.textui import progress
from django.utils.timezone import now
from django.core.management import call_command
from calaccess_raw.management import handle_command
from django.core.management.base import CommandError
from django.template.loader import render_to_string
from calaccess_raw.models.tracking import RawDataVersion
from calaccess_raw.management.commands import CalAccessCommand
from django.contrib.humanize.templatetags.humanize import naturaltime
from calaccess_raw.management.commands.downloadcalaccessrawdata import TestCommand as TestDownloadCommand
from calaccess_raw import (
    get_download_directory,
    get_test_download_directory,
    get_model_list
)
logger = logging.getLogger(__name__)


class Command(CalAccessCommand):
    """
    Download, unzip, clean and load the latest CAL-ACCESS database ZIP.
    """
    help = "Download, unzip, clean and load the latest CAL-ACCESS database ZIP"

    def add_arguments(self, parser):
        """
        Adds custom arguments specific to this command.
        """
        super(Command, self).add_arguments(parser)
        parser.add_argument(
            "--skip-download",
            action="store_false",
            dest="download",
            default=True,
            help="Skip downloading of the ZIP archive"
        )
        parser.add_argument(
            "--skip-clean",
            action="store_false",
            dest="clean",
            default=True,
            help="Skip cleaning up the raw data files"
        )
        parser.add_argument(
            "--skip-load",
            action="store_false",
            dest="load",
            default=True,
            help="Skip loading up the raw data files"
        )
        parser.add_argument(
            "--keep-files",
            action="store_true",
            dest="keep_files",
            default=False,
            help="Keep zip, unzipped, TSV and CSV files"
        )
        parser.add_argument(
            "--noinput",
            action="store_true",
            dest="noinput",
            default=False,
            help="Download the ZIP archive without asking permission"
        )
        parser.add_argument(
            "--test",
            "--use-test-data",
            action="store_true",
            dest="test_data",
            default=False,
            help="Use sampled test data (skips download, clean a load)"
        )
        parser.add_argument(
            "-a",
            "--app-name",
            dest="app_name",
            default="calaccess_raw",
            help="Name of Django app with models into which data will "
                 "be imported (if other not calaccess_raw)"
        )

    def handle(self, *args, **options):
        """
        Make it happen.
        """
        super(Command, self).handle(*args, **options)

        # set / compute any attributes that multiple class methods need
        self.app_name = options["app_name"]
        self.keep_files = options["keep_files"]
        self.test_mode = options['test_data']
        self.downloading = options['download']
        self.cleaning = options['clean']
        self.loading = options['load']
        self.noinput = options['noinput']

        if self.test_mode:
            # and always keep files when running test data
            self.keep_files = True
            self.data_dir = get_test_download_directory()
            # need to set this app-wide because cleancalaccessrawfile
            #   also calls get_download_directory
            settings.CALACCESS_DOWNLOAD_DIR = self.data_dir
            self.noinput = True
        else:
            self.data_dir = get_download_directory()

        os.path.exists(self.data_dir) or os.makedirs(self.data_dir)
        self.zip_path = os.path.join(self.data_dir, 'calaccess.zip')
        self.tsv_dir = os.path.join(self.data_dir, "tsv/")

        # Immediately check that the tsv directory exists when using test data,
        #   so we can stop immediately.
        if self.test_mode:
            if not os.path.exists(self.tsv_dir):
                raise CommandError("Data tsv directory does not exist "
                                   "at %s" % self.tsv_dir)
            elif self.verbosity:
                self.log("Using test data")

        self.csv_dir = os.path.join(self.data_dir, "csv/")
        os.path.exists(self.csv_dir) or os.makedirs(self.csv_dir)

        if self.test_mode:
            with open(self.data_dir + "/sampled_version.txt", "r") as f:
                current_release_datetime = f.readline()
                expected_size = f.readline()
        else:
            download_metadata = self.get_download_metadata()
            current_release_datetime = download_metadata['last-modified']
            expected_size = download_metadata['content-length']

        last_started_update = self.get_last_log()

        if self.test_mode:
            last_download = None
        else:
            try:
                last_download = self.command_logs.filter(
                    command='downloadcalaccessrawdata'
                ).order_by('-start_datetime')[0]
            except IndexError:
                last_download = None

        up_to_date = False
        can_resume = False

        # if there's a previously started update
        if last_started_update:
            # if current release datetime matches version of last started update
            if current_release_datetime == last_started_update.version.release_datetime:
                # if the last update finished
                if last_started_update.finish_datetime:
                    up_to_date = True
                else:
                    # if the last update didn't finish
                    # (but is still for the current version)
                    can_resume = True
            # if the last started update didn't finish
            elif not last_started_update.finish_datetime:
                # can resume update of old version as long as skipping download
                if not self.downloading:
                    can_resume = True
                # or if there is a last download
                elif last_download:
                    # and last download's version matches the outstanding update version
                    if last_download.version == last_started_update.version:
                        # and last download completed
                        if last_download.finish_datetime:
                            can_resume = True

        if self.noinput:
            # if already up-to-date, skip
            if up_to_date:
                self.log('Your database is already up-to-date.')
                logger.info("Your database is already up-to-date.")
                exit(0)
            # if not taking input and can resume, automatically go into resume mode
            self.resume_mode = can_resume
        else:
            prompt_context = dict(
                current_release_datetime=current_release_datetime,
                expected_size=size(expected_size),
                up_to_date=up_to_date,
                can_resume=can_resume,
            )

            last_finished_update = self.get_last_log(finished=True)

            if last_finished_update:
                loaded_v = last_finished_update.version
                prompt_context['since_loaded_version'] = naturaltime(loaded_v.release_datetime)
            else:
                prompt_context['since_loaded_version'] = None

            prompt = render_to_string(
                'calaccess_raw/updatecalaccessrawdata.txt',
                prompt_context,
            )

            if can_resume:
                if self.confirm_proceed(prompt):
                    self.resume_mode = True
                else:
                    self.resume_mode = False
                    if not self.confirm_proceed('Do you want re-start your update?\n'):
                        raise CommandError("Update cancelled")
            else:
                self.resume_mode = False
                if not self.confirm_proceed(prompt):
                    raise CommandError("Update cancelled")

        if self.resume_mode:
            self.log_record = last_started_update
        else:
            # get or create a version
            # .get_or_create() throws IntegrityError
            try:
                version = self.raw_data_versions.get(
                    release_datetime=current_release_datetime
                )
            except RawDataVersion.DoesNotExist:
                version = self.raw_data_versions.create(
                    release_datetime=current_release_datetime,
                    size=expected_size
                )
            # create a new log record
            self.log_record = self.command_logs.create(
                version=version,
                command=self,
                called_by=self.get_caller_log()
            )

        # if the user could have resumed but didn't
        force_restart_download = can_resume and not self.resume_mode

        # if not skipping download, and there's a previous download
        if self.downloading and last_download:
            # if not forcing a restart
            if not force_restart_download:
                # check if version we are updating is last one being downloaded
                if self.log_record.version == last_download.version:
                    # if it finished
                    if last_download.finish_datetime:
                        self.log('Already downloaded.')
                        self.downloading = False

        if self.downloading:
            if self.test_mode:
                handle_command(TestDownloadCommand, verbosity=self.verbosity)
            else:
                call_command(
                    "downloadcalaccessrawdata",
                    keep_files=self.keep_files,
                    verbosity=self.verbosity,
                    noinput=True,
                    restart=force_restart_download,
                )
            if self.verbosity:
                self.duration()

        # execute the other steps that haven't been skipped
        if options['clean']:
            self.clean()
            if self.verbosity:
                self.duration()

        if options['load']:
            self.load()
            if self.verbosity:
                self.duration()

        self.log_record.finish_datetime = now()
        self.log_record.save()

        if self.verbosity:
            self.success("Done!")
        logger.info("Done!")

    def clean(self):
        """
        Clean up the raw data files from the state so they are ready to get loaded into the database.
        """
        if self.verbosity:
            self.header("Cleaning data files")

        tsv_list = os.listdir(self.tsv_dir)

        if self.resume_mode:
            # get finished clean command logs of last update
            prev_cleaned = [
                x.file_name + '.TSV'
                for x in self.log_record.called.filter(
                    command='cleancalaccessrawfile',
                    finish_datetime__isnull=False
                )
            ]
            self.log("{} files already cleaned.".format(len(prev_cleaned)))
            # remove these from tsv_list
            tsv_list = [x for x in tsv_list if x not in prev_cleaned]

        # Loop through all the files in the source directory
        if self.verbosity:
            tsv_list = progress.bar(tsv_list)
        for name in tsv_list:
            call_command(
                "cleancalaccessrawfile",
                name,
                verbosity=self.verbosity,
                keep_files=self.keep_files,
            )

    def load(self):
        """
        Loads the cleaned up csv files into the database.
        """
        if self.verbosity:
            self.header("Loading data files")

        model_list = [
            x for x in get_model_list() if os.path.exists(x.objects.get_csv_path())
        ]

        if self.resume_mode:
            # get finished load command logs of last update
            prev_loaded = [
                x.file_name
                for x in self.log_record.called.filter(
                    command='loadcalaccessrawfile',
                    finish_datetime__isnull=False
                )
            ]
            self.log("{} models already loaded.".format(len(prev_loaded)))
            # remove these from model_list
            model_list = [x for x in model_list if x._meta.db_table not in prev_loaded]

        if self.verbosity:
            model_list = progress.bar(model_list)
        for model in model_list:
            call_command(
                "loadcalaccessrawfile",
                model.__name__,
                verbosity=self.verbosity,
                keep_files=self.keep_files,
                app_name=self.app_name,
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Clean a source CAL-ACCESS TSV file and reformat it as a CSV.
"""
from __future__ import unicode_literals
import os
import csv
from io import StringIO
from django.conf import settings
from django.core.files import File
from django.utils import six
from django.utils.timezone import now
from csvkit import CSVKitReader, CSVKitWriter
from calaccess_raw import get_download_directory
from calaccess_raw.management.commands import CalAccessCommand
from calaccess_raw.models.tracking import RawDataVersion


class Command(CalAccessCommand):
    """
    Clean a source CAL-ACCESS TSV file and reformat it as a CSV.
    """
    help = 'Clean a source CAL-ACCESS TSV file and reformat it as a CSV'

    def add_arguments(self, parser):
        """
        Adds custom arguments specific to this command.
        """
        super(Command, self).add_arguments(parser)
        parser.add_argument(
            'file_name',
            help="Name of the TSV file to be cleaned and discarded for a CSV"
        )
        parser.add_argument(
            "--keep-files",
            action="store_true",
            dest="keep_files",
            default=False,
            help="Keep original TSV file"
        )

    def handle(self, *args, **options):
        """
        Make it happen.
        """
        super(Command, self).handle(*args, **options)

        # Set options
        self.file_name = options['file_name']
        self.data_dir = get_download_directory()
        self.tsv_dir = os.path.join(self.data_dir, "tsv/")
        self.csv_dir = os.path.join(self.data_dir, "csv/")
        self.log_dir = os.path.join(self.data_dir, "log/")
        self.error_log_path = os.path.join(
            self.log_dir,
            self.file_name.lower().replace("tsv", "errors.csv")
        )

        if self.verbosity > 2:
            self.log(" Cleaning %s" % self.file_name)

        caller = self.get_caller_log()

        if caller:
            # if called by another command, use its version record
            self.version = caller.version
            self.log_record = self.command_logs.create(
                version=self.version,
                command=self,
                called_by=caller,
                file_name=self.file_name.upper().replace('.TSV', '')
            )
        else:
            # try getting the most recent version
            try:
                self.version = self.raw_data_versions.latest('release_datetime')
            except RawDataVersion.DoesNotExist:
                # if there's no version, assume this is a test and do not log
                # TODO: Figure out a more direct way to handle this
                self.version = None
            else:
                self.log_record = self.command_logs.create(
                    # if called by another command, use it's version
                    version=self.version,
                    command=self,
                    file_name=self.file_name.upper().replace('.TSV', '')
                )

        self.clean()

        # unless keeping files, remove tsv files
        if not options['keep_files']:
            os.remove(os.path.join(self.tsv_dir, options['file_name']))

        # save the log record
        self.log_record.finish_datetime = now()
        self.log_record.save()

    def clean(self):
        """
        Cleans the provided source TSV file and writes it out in CSV format.
        """
        # Up the CSV data limit
        csv.field_size_limit(1000000000)

        # Input and output paths
        tsv_path = os.path.join(self.tsv_dir, self.file_name)
        csv_path = os.path.join(
            self.csv_dir,
            self.file_name.lower().replace("tsv", "csv")
        )

        # Reader
        tsv_file = open(tsv_path, 'rbU')

        # Writer
        csv_file = open(csv_path, 'w')
        csv_writer = CSVKitWriter(csv_file, quoting=csv.QUOTE_ALL)

        # Pull and clean the headers
        try:
            headers = tsv_file.readline()
        except StopIteration:
            return
        headers = headers.decode("ascii", "replace")
        headers_csv = CSVKitReader(StringIO(headers), delimiter=str('\t'))
        try:
            headers_list = next(headers_csv)
        except StopIteration:
            return
        headers_count = len(headers_list)
        csv_writer.writerow(headers_list)

        log_rows = []

        # Loop through the rest of the data
        line_number = 1
        for tsv_line in tsv_file:
            line_number += 1

            # Log empty lines, then skip
            if tsv_line == '\n':
                if self.verbosity > 2:
                    msg = '  Line %s is empty'
                    self.failure(msg % (
                        line_number,
                    ))
                continue

            # Goofing around with the encoding while we're in there.
            tsv_line = tsv_line.decode("ascii", "replace")
            if six.PY2:
                tsv_line = tsv_line.replace('\ufffd', '?')

            # Nuke any null bytes
            null_bytes = tsv_line.count('\x00')
            if null_bytes:
                tsv_line = tsv_line.replace('\x00', ' ')

            # Nuke ASCII 26 char, the "substitute character"
            # or chr(26) in python
            sub_char = tsv_line.count('\x1a')
            if sub_char:
                tsv_line = tsv_line.replace('\x1a', '')

            # Remove any extra newline chars
            tsv_line = tsv_line.replace("\r\n", "").replace("\r", "").replace("\n", "")

            # Split on tabs so we can later spit it back out as CSV
            csv_field_list = tsv_line.split("\t")

            # Check if our values line up with our headers
            # and if not, see if CSVkit can sort out the problems
            if not len(csv_field_list) == headers_count:
                csv_field_list = next(CSVKitReader(
                    StringIO(tsv_line),
                    delimiter=str('\t')
                ))

                if not len(csv_field_list) == headers_count:
                    if self.verbosity > 2:
                        msg = '  Bad parse of line %s (%s headers, %s values)'
                        self.failure(msg % (
                            line_number,
                            len(headers_list),
                            len(csv_field_list)
                        ))
                    log_rows.append([
                        line_number,
                        len(headers_list),
                        len(csv_field_list),
                        ','.join(csv_field_list)
                    ])
                    continue

            # Write out the row
            csv_writer.writerow(csv_field_list)

        # Log errors if there are any
        if log_rows:
            if self.verbosity > 1:
                msg = '  %s errors logged (not including empty lines)'
                self.failure(msg % (len(log_rows)))
            self.log_errors(log_rows)

        raw_file = self.raw_data_files.get_or_create(
            version=self.version,
            file_name=self.log_record.file_name
        )[0]

        # Add counts to raw_file_record
        raw_file.download_columns_count = headers_count
        raw_file.download_records_count = line_number - 1
        raw_file.clean_columns_count = headers_count
        raw_file.clean_records_count = line_number - 1 - len(log_rows)
        raw_file.error_count = len(log_rows)

        # Add file size to the raw_file_record
        raw_file.download_file_size = os.path.getsize(tsv_path) or 0
        raw_file.clean_file_size = os.path.getsize(csv_path) or 0

        # Save it
        raw_file.save()

        # Shut it down
        tsv_file.close()
        csv_file.close()

        if getattr(settings, 'CALACCESS_STORE_ARCHIVE', False):
            if self.verbosity > 2:
                self.log(" Archiving {0}.csv".format(raw_file.file_name))
            # Remove previous .CSV and error log files
            raw_file.clean_file_archive.delete()
            raw_file.error_log_archive.delete()

            # Open up the .CSV file so we can wrap it in the Django File obj
            with open(csv_path) as csv_file:
                # Save the .CSV on the raw data file
                raw_file.clean_file_archive.save(
                    self.file_name.lower().replace("tsv", "csv"),
                    File(csv_file),
                )
            # if there are any errors, archive the log too
            if log_rows:
                if self.verbosity > 2:
                    self.log(" Archiving {0}.errors.csv".format(raw_file.file_name))
                with open(self.error_log_path) as error_file:
                    raw_file.error_log_archive.save(
                        os.path.basename(self.error_log_path),
                        File(error_file),
                    )

    def log_errors(self, rows):
        """
        Log any errors to a CSV file.
        """
        # Make sure the log directory exists
        os.path.exists(self.log_dir) or os.makedirs(self.log_dir)

        # Log writer
        log_file = open(self.error_log_path, 'w')
        log_writer = CSVKitWriter(log_file, quoting=csv.QUOTE_ALL)

        # Add the headers
        log_writer.writerow([
            'Line number',
            'Headers len',
            'Fields len',
            'Line value'
        ])

        # Log out the rows
        log_writer.writerows(rows)

        # Shut it down
        log_file.close()
